/*

  This file is a part of GRALE, a library to facilitate the simulation
  and inversion of gravitational lenses.

  Copyright (C) 2008-2012 Jori Liesenborgs

  Contact: jori.liesenborgs@gmail.com
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
*/

/**
 * \file cosmology.h
 */

#ifndef GRALE_COSMOLOGY_H

#define GRALE_COSMOLOGY_H

#include "graleconfig.h"
#include <errut/errorbase.h>

namespace grale
{

/** This class can be used to calculate angular diameter distances
 *  in a specific kind of cosmology.
 */
class GRALE_IMPORTEXPORT Cosmology : public errut::ErrorBase
{
public:
	/** Create an instance to calculate angular diameter distances in a specific
	 *  kind of cosmological model.
	 *  Create an instance to calculate angular diameter distances in a specific
	 *  kind of cosmological model.
	 *  \param h Speficies the Hubble constant: H_0 = h * 100 km s^-1 Mpc^-1
	 *  \param Wm Omega_m
	 *  \param Wr Omega_r
	 *  \param Wv Omega_v
	 *  \param w Equation of state parameter for dark matter
	 */
	Cosmology(double h = 0.71, double Wm = 0.27, double Wr = 0, double Wv = 0.73, double w = -1.0);
	~Cosmology();

	/** Sets the Hubble constant to H_0 = h * 100 km s^-1 Mpc^-1. */
	void setH(double h)								{ m_h = h; }

	/** Sets Omega_m to \c Wm. */
	void setMatter(double Wm)							{ m_Wm = Wm; }

	/** Sets Omega_r to \c Wr. */
	void setRadiation(double Wr)							{ m_Wr = Wr; }

	/** Sets Omega_v to \c Wv. */
	void setVacuum(double Wv)							{ m_Wv = Wv; }

	/** Sets the equation of state parameter for dark matter to \c w. */
	void setVacuumEOSParameter(double w)						{ m_w = w; }

	double getH() const								{ return m_h; }
	double getMatter() const							{ return m_Wm; }
	double getRadiation() const							{ return m_Wr; }
	double getVacuum() const							{ return m_Wv; }
	double getVacuumEOSParameter() const						{ return m_w; }

	/** Returns the angular diameter distance between redshifts \c z1 and \c z2.
	 *  Returns the angular diameter distance between redshifts \c z1 and \c z2.
	 *  The precision of the numerical integration is controlled by the value
	 *  of \c integrationSteps.
	 */
	double getAngularDiameterDistance(double z1, double z2, int integrationSteps = 1000) const;
	
	/** Returns the angular diameter distance corresponding to \c z.
	 *  Returns the angular diameter distance corresponding to \c z.
	 *  The precision of the numerical integration is controlled by the value
	 *  of \c integrationSteps.
	 */
	double getAngularDiameterDistance(double z, int integrationSteps = 1000) const	{ return getAngularDiameterDistance(0, z, integrationSteps); }
private:
	double m_h;
	double m_Wm, m_Wr, m_Wv;
	double m_w;
};
	
} // end namespace

#endif // GRALE_COSMOLOGY_H

