/*

  This file is a part of GRALE, a library to facilitate the simulation
  and inversion of gravitational lenses.

  Copyright (C) 2008-2012 Jori Liesenborgs

  Contact: jori.liesenborgs@gmail.com
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
*/

#include "graleconfig.h"
#include "gridlensinversiongafactoryparams.h"
#include "imagesdata.h"
#include "gravitationallens.h"
#include <serut/memoryserializer.h>
#include <serut/dummyserializer.h>
#include <vector>

#include "debugnew.h"

namespace grale
{

GridLensInversionGAFactoryParams::GridLensInversionGAFactoryParams() : LensInversionGAFactoryParams(LensInversionGAFactoryParams::GridInversion)
{
	deleteimages = false;
	D_d = 0;
	z_d = 0;
	maxgenerations = 0;
	usemassweights = false;
	basisfunction = PlummerBasis;
	m_allowNegative = false;
	m_pBaseLens = 0;
}

GridLensInversionGAFactoryParams::GridLensInversionGAFactoryParams(int maxgen, const std::list<ImagesDataExtended *> &images, 
		                                                   const std::list<GridSquare> &gridsquares,
								   double D_d, double z_d, double massscale, bool copyimages, bool useweights,
								   BasisFunctionType b, bool allowNegativeValues,
								   const GravitationalLens *pBaseLens,
								   MassSheetSearchType sheetSearchType) : LensInversionGAFactoryParams(LensInversionGAFactoryParams::GridInversion)
{
	if (copyimages)
		deleteimages = true;
	else
		deleteimages = false;
	
	maxgenerations = maxgen;
	usemassweights = useweights; 

	GridLensInversionGAFactoryParams::D_d = D_d;
	GridLensInversionGAFactoryParams::z_d = z_d;
	GridLensInversionGAFactoryParams::massscale = massscale; 
	
	std::list<ImagesDataExtended *>::const_iterator it;

	for (it = images.begin() ; it != images.end() ; it++)
	{
		ImagesDataExtended *img = (*it);
		ImagesDataExtended *img2;

		if (copyimages)
			img2 = new ImagesDataExtended(*img);
		else
			img2 = img;
			
		GridLensInversionGAFactoryParams::images.push_back(img2);
	}

	std::list<GridSquare>::const_iterator it2;

	for (it2 = gridsquares.begin() ; it2 != gridsquares.end() ; it2++)
		GridLensInversionGAFactoryParams::gridsquares.push_back(*it2);

	basisfunction = b;

	m_allowNegative = allowNegativeValues;
	m_pBaseLens = 0;

	if (pBaseLens)
	{
		serut::DummySerializer dumSer;

		if (pBaseLens->write(dumSer))
		{
			int numBytes = dumSer.getBytesWritten();
			std::vector<uint8_t> buffer(numBytes);
			serut::MemorySerializer memSer(&(buffer[0]), numBytes, &(buffer[0]), numBytes);
			std::string errStr;
			
			pBaseLens->write(memSer);

			GravitationalLens::read(memSer, &m_pBaseLens, errStr);
		}
	}

	m_massSheetSearchType = sheetSearchType;
}

GridLensInversionGAFactoryParams::~GridLensInversionGAFactoryParams()
{
	if (deleteimages)
	{
		while (!images.empty())
		{
			delete (*(images.begin()));
			images.pop_front();
		}
	}

	if (m_pBaseLens)
		delete m_pBaseLens;
}

#define BASISFUNCTION_PLUMMER 1
#define BASISFUNCTION_SQUARE  2
#define BASISFUNCTION_GAUSS   3

#define SHEETSEARCHTYPE_NONE  	0
#define SHEETSEARCHTYPE_GENOME	1
#define SHEETSEARCHTYPE_LOOP	2

bool GridLensInversionGAFactoryParams::write(serut::SerializationInterface &si) const
{
	if (!si.writeInt32(maxgenerations))
	{
		setErrorString(si.getErrorString());
		return false;
	}
	if (!si.writeDouble(D_d))
	{
		setErrorString(si.getErrorString());
		return false;
	}
	if (!si.writeDouble(z_d))
	{
		setErrorString(si.getErrorString());
		return false;
	}
	if (!si.writeDouble(massscale))
	{
		setErrorString(si.getErrorString());
		return false;
	}

	int32_t sizes[2];

	sizes[0] = images.size();
	sizes[1] = gridsquares.size();

	if (!si.writeInt32s(sizes,2))
	{
		setErrorString(si.getErrorString());
		return false;
	}

	std::list<ImagesDataExtended *>::const_iterator it;

	for (it = images.begin() ; it != images.end() ; it++)
	{
		if (!(*it)->write(si))
		{
			setErrorString((*it)->getErrorString());
			return false;
		}
	}

	std::list<GridSquare>::const_iterator it2;

	for (it2 = gridsquares.begin() ; it2 != gridsquares.end() ; it2++)
	{
		double params[3];

		params[0] = (*it2).getSize();
		params[1] = (*it2).getCenter().getX();
		params[2] = (*it2).getCenter().getY();
	
		if (!si.writeDoubles(params,3))
		{
			setErrorString(si.getErrorString());
			return false;
		}
	}

	int32_t w = (usemassweights)?1:0;
	if (!si.writeInt32(w))
	{
		setErrorString(si.getErrorString());
		return false;
	}

	int32_t b = 0;

	switch(basisfunction)
	{
	case PlummerBasis:
		b = BASISFUNCTION_PLUMMER;
		break;
	case SquareBasis:
		b = BASISFUNCTION_SQUARE;
		break;
	case GaussBasis:
		b = BASISFUNCTION_GAUSS;
		break;
	default:
		setErrorString("Detected invalid basis function setting");
		return false;
	}

	if (!si.writeInt32(b))
	{
		setErrorString(si.getErrorString());
		return false;
	}

	int32_t n = (m_allowNegative)?1:0;

	if (!si.writeInt32(n))
	{
		setErrorString(si.getErrorString());
		return false;
	}

	int32_t val = 0;
	
	if (m_pBaseLens)
		val = 1;
	
	if (!si.writeInt32(val))
	{
		setErrorString(si.getErrorString());
		return false;
	}

	if (m_pBaseLens)
	{
		if (!m_pBaseLens->write(si))
		{
			setErrorString(m_pBaseLens->getErrorString());
			return false;
		}
	}

	switch(m_massSheetSearchType)
	{
	case Loop:
		val = SHEETSEARCHTYPE_LOOP;
		break;
	case Genome:
		val = SHEETSEARCHTYPE_GENOME;
		break;
	case None:
	default:
		val = SHEETSEARCHTYPE_NONE;
	}

	if (!si.writeInt32(val))
	{
		setErrorString(si.getErrorString());
		return false;
	}

	return true;
}

bool GridLensInversionGAFactoryParams::read(serut::SerializationInterface &si)
{
	if (deleteimages)
	{
		while (!images.empty())
		{
			delete (*(images.begin()));
			images.pop_front();
		}
	}
	gridsquares.clear();
	images.clear();

	if (m_pBaseLens)
	{
		delete m_pBaseLens;
		m_pBaseLens = 0;
	}
	
	deleteimages = true;
	
	int32_t maxgen;

	if (!si.readInt32(&maxgen))
	{
		setErrorString(si.getErrorString());
		return false;
	}
	maxgenerations = (int)maxgen;
	
	if (!si.readDouble(&D_d))
	{
		setErrorString(si.getErrorString());
		return false;
	}
	if (!si.readDouble(&z_d))
	{
		setErrorString(si.getErrorString());
		return false;
	}
	if (!si.readDouble(&massscale))
	{
		setErrorString(si.getErrorString());
		return false;
	}

	int32_t sizes[2];

	if (!si.readInt32s(sizes,2))
	{
		setErrorString(si.getErrorString());
		return false;
	}

	int32_t numimages = sizes[0];
	int32_t numsquares = sizes[1];
	
	for (int32_t i = 0 ; i < numimages ; i++)
	{
		ImagesDataExtended *img = new ImagesDataExtended();
		if (!img->read(si))
		{
			setErrorString(img->getErrorString());
			delete img;
			return false;
		}
		images.push_back(img);
	}

	for (int32_t i = 0 ; i < numsquares ; i++)
	{
		double info[3];

		if (!si.readDoubles(info,3))
		{
			setErrorString(si.getErrorString());
			return false;
		}
		gridsquares.push_back(GridSquare(Vector2D<double>(info[1],info[2]),info[0]));
	}
	
	int32_t w;

	if (!si.readInt32(&w))
	{
		setErrorString(si.getErrorString());
		return false;
	}
	if (w == 0)
		usemassweights = false;
	else
		usemassweights = true;

	int32_t b;

	if (!si.readInt32(&b))
	{
		setErrorString(si.getErrorString());
		return false;
	}

	switch(b)
	{
	case BASISFUNCTION_PLUMMER:
		basisfunction = PlummerBasis;
		break;
	case BASISFUNCTION_SQUARE:
		basisfunction = SquareBasis;
		break;
	case BASISFUNCTION_GAUSS:
		basisfunction = GaussBasis;
		break;
	default:
		setErrorString("Read invalid basis function identifier");
		return false;
	}

	int32_t n;

	if (!si.readInt32(&n))
	{
		setErrorString(si.getErrorString());
		return false;
	}

	if (n == 0)
		m_allowNegative = false;
	else 
		m_allowNegative = true;

	int32_t val;

	if (!si.readInt32(&val))
	{
		setErrorString(si.getErrorString());
		return false;
	}

	if (val != 0)
	{
		std::string errStr;

		if (!GravitationalLens::read(si, &m_pBaseLens, errStr))
		{
			setErrorString(errStr);
			return false;
		}
	}

	if (!si.readInt32(&val))
	{
		setErrorString(si.getErrorString());
		return false;
	}

	switch(val)
	{
	case SHEETSEARCHTYPE_NONE:
		m_massSheetSearchType = None;
		break;
	case SHEETSEARCHTYPE_GENOME:
		m_massSheetSearchType = Genome;
		break;
	case SHEETSEARCHTYPE_LOOP:
		m_massSheetSearchType = Loop;
		break;
	default:
		setErrorString("Invalid mass-sheet search type");
		return false;
	}
	
	return true;
}
	
} // end namespace

