/*

  This file is a part of GRALE, a library to facilitate the simulation
  and inversion of gravitational lenses.

  Copyright (C) 2008-2012 Jori Liesenborgs

  Contact: jori.liesenborgs@gmail.com
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
*/

#include "graleconfig.h"
#include "treelensinversiongafactorybase.h"
#include "treelensinversiongafactoryparams.h"
#include "treelensinversiongenomebase.h"
#include "plummerbackprojecttree.h"
#include "gridtree.h"

#include <iostream>

#ifdef SHOWEVOLUTION
#include "gravitationallens.h"
#include <sys/stat.h>
#include <sys/types.h>
#endif // SHOWEVOLUTION

#include "debugnew.h"

namespace grale
{

TreeLensInversionGAFactoryBase::TreeLensInversionGAFactoryBase()
{
	m_pBackProjectTree = 0;
	m_pCurrentParams = 0;
	m_maxGenerations = 0;
	m_maxDepth = 0;
	m_maxBasisFunctions = 0;
	m_useWeights = false;
}

TreeLensInversionGAFactoryBase::~TreeLensInversionGAFactoryBase()
{
	if (m_pBackProjectTree)
		delete m_pBackProjectTree;
	if (m_pCurrentParams)
		delete m_pCurrentParams;
}

mogal::GAFactoryParams *TreeLensInversionGAFactoryBase::createParamsInstance() const
{
	return new TreeLensInversionGAFactoryParams();
}

const mogal::GAFactoryParams *TreeLensInversionGAFactoryBase::getCurrentParameters() const
{
	return m_pCurrentParams;
}

bool TreeLensInversionGAFactoryBase::init(const mogal::GAFactoryParams *pParams)
{
	if (m_pCurrentParams != 0)
	{
		setErrorString("Already initialized");
		return false;
	}

	if (pParams == 0)
	{
		setErrorString("Specified parameters can't be null");
		return false;
	}
	
	const LensInversionGAFactoryParams *pParams1 = (const LensInversionGAFactoryParams *)pParams;
	if (pParams1->getInversionType() != LensInversionGAFactoryParams::TreeInversion)
	{
		setErrorString("This algorithm is a tree based one");
		return false;
	}

	const TreeLensInversionGAFactoryParams *pParams2 = (const TreeLensInversionGAFactoryParams *)pParams;
	
	m_pCurrentParams = new TreeLensInversionGAFactoryParams(pParams2->getMaximumNumberOfGenerations(), 
			                                        pParams2->getImages(), 
								pParams2->getCenter(), 
								pParams2->getAreaSize(), 
								pParams2->getMaxTreeDepth(), 
								pParams2->getMaxBasisFunctions(), 
								pParams2->getD_d(), 
								pParams2->getMassScale(), 
								true, 
								pParams2->getWidthScaleFactor(),
								pParams2->useWeights(),
								pParams2->getInitialSubdivision(),
								pParams2->getBasisFunctionType());

	int maxBasisFunctions = pParams2->getMaxBasisFunctions();

	m_maxGenomeSize = ((MAX(sizeof(float),sizeof(int32_t))+sizeof(int32_t))*(maxBasisFunctions+10)*4)/3;

	std::list<ImagesDataExtended *> imageList;
	
	for (std::list<ImagesDataExtended *>::const_iterator it = pParams2->getImages().begin() ; it != pParams2->getImages().end() ; it++)
		imageList.push_back(*it);
	
	// perform sub-initialization
	if (!subInit(imageList)) // imagelist can be modified if not everything needs to be projected back
	{
		delete m_pCurrentParams;
		m_pCurrentParams = 0;
		return false;
	}

	if (m_pCurrentParams->getBasisFunctionType() == TreeLensInversionGAFactoryParams::PlummerBasis)
	{
		PlummerBackProjectTree *pBackProjectTree = new PlummerBackProjectTree();
		
		pBackProjectTree->init(m_pCurrentParams->getD_d(), 
				       m_pCurrentParams->getMassScale(), 
				       m_pCurrentParams->getWidthScaleFactor(),
				       imageList,
				       m_pCurrentParams->getCenter(),
				       m_pCurrentParams->getAreaSize(),
				       m_pCurrentParams->getInitialSubdivision(),
				       m_pCurrentParams->getMaxTreeDepth(),
				       m_pCurrentParams->useWeights());
		
		m_pBackProjectTree = pBackProjectTree;
	}
	else // unknown
	{
		setErrorString("Unknown basis function type specified");
		delete m_pCurrentParams;
		m_pCurrentParams = 0;
		return false;
	}

	m_maxGenerations = m_pCurrentParams->getMaximumNumberOfGenerations();
	m_maxDepth = m_pCurrentParams->getMaxTreeDepth();
	m_maxBasisFunctions = m_pCurrentParams->getMaxBasisFunctions();
	m_useWeights = m_pCurrentParams->useWeights();
	m_firstLevelSubdiv = m_pCurrentParams->getInitialSubdivision();
	
	return true;
}

mogal::Genome *TreeLensInversionGAFactoryBase::createNewGenome() const
{
	return allocateNewGenome((TreeLensInversionGAFactoryBase *)this, m_pBackProjectTree, 0, m_firstLevelSubdiv);
}

bool TreeLensInversionGAFactoryBase::writeGenome(serut::SerializationInterface &si, const mogal::Genome *g) const
{
	const TreeLensInversionGenomeBase *pGenome = (const TreeLensInversionGenomeBase *)g;
	const GridNodeBase *pRootNode = pGenome->getRootNode();

	if (!writeNode(si, pRootNode)) // error string is set in this function
		return false;
	
	return true;
}

#define NODETYPE_WEIGHTNODE	1
#define NODETYPE_TREENODE	2

bool TreeLensInversionGAFactoryBase::writeNode(serut::SerializationInterface &si, const GridNodeBase *pNode) const
{
	if (pNode->getNodeType() == GridNodeBase::TreeNode)
	{
		const GridTreeNode *pTreeNode = (const GridTreeNode *)pNode;

		if (!si.writeInt32(NODETYPE_TREENODE))
		{
			setErrorString(std::string("Couldn't write tree node identifier: ") + si.getErrorString());
			return false;
		}

		int numChildren = pTreeNode->getNumberOfChildren();
		
		if (!si.writeInt32(numChildren))
		{
			setErrorString(std::string("Couldn't write number of children: ") + si.getErrorString());
			return false;
		}
		
		for (int i = 0 ; i < numChildren ; i++)
		{
			const GridNodeBase *pChildNode = pTreeNode->getChild(i);

			if (!writeNode(si, pChildNode))
				return false;
		}
	}
	else // weight node
	{
		const GridWeightNode *pWeightNode = (const GridWeightNode *)pNode;
		float weight = pWeightNode->getWeight();

		if (!si.writeInt32(NODETYPE_WEIGHTNODE))
		{
			setErrorString(std::string("Couldn't write weight node identifier: ") + si.getErrorString());
			return false;
		}
	
		if (!si.writeFloat(weight))
		{
			setErrorString(std::string("Couldn't write weight: ") + si.getErrorString());
			return false;
		}
	}
	return true;
}

bool TreeLensInversionGAFactoryBase::readGenome(serut::SerializationInterface &si, mogal::Genome **g) const
{
	GridNodeBase *pRootNode = 0;

	if (!readNode(si, &pRootNode, 0))
		return false;

	*g = allocateNewGenome((TreeLensInversionGAFactoryBase *)this, m_pBackProjectTree, pRootNode, -1);
	return true;
}

bool TreeLensInversionGAFactoryBase::readNode(serut::SerializationInterface &si, GridNodeBase **pNode, int level) const
{
	int32_t id;

	if (!si.readInt32(&id))
	{
		setErrorString(std::string("Couldn't read node identifier: ") + si.getErrorString());
		return false;
	}

	if (id == NODETYPE_TREENODE)
	{
		int32_t numChildren = 0;

		if (!si.readInt32(&numChildren))
		{
			setErrorString(std::string("Couldn't read number of children: ") + si.getErrorString());
			return false;
		}
		
		GridTreeNode *pTreeNode = new GridTreeNode(0, -1, level, numChildren);

		for (int32_t i = 0 ; i < numChildren ; i++)
		{
			GridNodeBase *pChild = 0;

			if (!readNode(si, &pChild, level+1))
			{
				delete pTreeNode;
				return false;
			}
			pTreeNode->setNodeNoLevel(i, pChild);
		}

		*pNode = pTreeNode;
	}
	else if (id == NODETYPE_WEIGHTNODE)
	{
		float weight;

		if (!si.readFloat(&weight))
		{
			setErrorString(std::string("Couldn't read node weight: ") +  si.getErrorString());
			return false;
		}

		GridWeightNode *pWeightNode = new GridWeightNode(weight, 0, -1, level);

		*pNode = pWeightNode;
	}
	else
	{
		setErrorString("Read invalid node identifier");
		return false;
	}
	return true;
}

bool TreeLensInversionGAFactoryBase::writeGenomeFitness(serut::SerializationInterface &si, const mogal::Genome *g) const
{
	const TreeLensInversionGenomeBase *g2 = (const TreeLensInversionGenomeBase *)g;
	int num = getNumberOfFitnessComponents();
	const float *f = g2->getFitnessValues();

	if (!si.writeFloats(f, num))
	{
		setErrorString(si.getErrorString());
		return false;
	}
	if (!si.writeFloat(g2->getScaleFactor()))
	{
		setErrorString(si.getErrorString());
		return false;
	}
	return true;
}

bool TreeLensInversionGAFactoryBase::readGenomeFitness(serut::SerializationInterface &si, mogal::Genome *g) const
{
	TreeLensInversionGenomeBase *g2 = (TreeLensInversionGenomeBase *)g;
	int num = getNumberOfFitnessComponents();
	float x[TREELENSINVERSIONGENOMEBASE_MAXFITNESSCOMP+1];

	if (!si.readFloats(x, num+1))
	{
		setErrorString(si.getErrorString());
		return false;
	}
	g2->setFitnessValues(x);
	g2->setScaleFactor(x[num]);
	return true;
}

bool TreeLensInversionGAFactoryBase::writeCommonGenerationInfo(serut::SerializationInterface &si) const
{
	return true;
}

bool TreeLensInversionGAFactoryBase::readCommonGenerationInfo(serut::SerializationInterface &si)
{
	return true;
}

#ifdef SHOWEVOLUTION

void TreeLensInversionGAFactoryBase::onSortedPopulation(const std::vector<mogal::GenomeWrapper> &population)
{
	char str[1024];
	int generation = getCurrentAlgorithm()->getCurrentGeneration();

	sprintf(str,"g_%d",generation);
	if(mkdir(str,S_IRWXU) != 0)
		return;
	
	for (int i = 0 ; i < (int)population.size() ; i++)
	{
		char str[1024];
		sprintf(str,"g_%d/p_%d_%d_%d.lensdata",generation,i,population[i].getParent1(),population[i].getParent2());

		const LensInversionGenomeBase *g = (const LensInversionGenomeBase *)population[i].getGenome();

		double m = 0;
		std::string errstr;
		GravitationalLens *lens = g->createLens(&m,errstr);
		lens->save(str);
		delete lens;
	}
}

#endif // SHOWEVOLUTION

} // end namespace
