/*

  This file is a part of GRALE, a library to facilitate the simulation
  and inversion of gravitational lenses.

  Copyright (C) 2008-2012 Jori Liesenborgs

  Contact: jori.liesenborgs@gmail.com
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
*/

#include "treelensinversiongenomebase.h"
#include "treelensinversiongafactorybase.h"
#include "treelensinversiongafactoryparams.h"
#include "multipleplummerlens.h"
#include "multiplesquarelens.h"
#include "multiplegausslens.h"
#include "doublepeakdistribution.h"
#include "simpleuniformdistribution.h"
#include "constants.h"
#include "gridtree.h"
#include "gaussiandistribution.h"

#include <iostream>

#include "debugnew.h"


namespace grale
{

TreeLensInversionGenomeBase::TreeLensInversionGenomeBase(TreeLensInversionGAFactoryBase *pFactory, 
						         BackProjectTreeBase *pBackProjectTree,
							 int firstLevelSubdiv)
{
	m_pFactory = pFactory;
	m_pBackProjectTree = pBackProjectTree;
	m_activeFitnessComponent = 0;
	
	for (int i = 0 ; i < TREELENSINVERSIONGENOMEBASE_MAXFITNESSCOMP ; i++)
		m_fitnessValues[i] = 0;
	
	m_scaleFactor = 1.0f;

	// Create an initial root node
	int numChildren = firstLevelSubdiv*firstLevelSubdiv;
	GridTreeNode *pRootNode = new GridTreeNode(0, -1, 0, numChildren);
	SimpleUniformDistribution uniDist(m_pFactory->getRandomNumberGenerator());

	for (int i = 0 ; i < numChildren ; i++)
		pRootNode->setNode(i, new GridWeightNode((float)uniDist.pickNumber(), 0, -1, 0));

//	std::cout << "firstLevelSubdiv = " << firstLevelSubdiv << std::endl;
//	std::cout << "New genome: " << std::endl;
//	pRootNode->print();

	m_pRootNode = pRootNode;
	rescale();
}

TreeLensInversionGenomeBase::TreeLensInversionGenomeBase(TreeLensInversionGAFactoryBase *pFactory, 
		                                         BackProjectTreeBase *pBackProjectTree,
							 GridNodeBase *pRootNode)
{
	m_pFactory = pFactory;
	m_pBackProjectTree = pBackProjectTree;
	m_activeFitnessComponent = 0;
	
	for (int i = 0 ; i < TREELENSINVERSIONGENOMEBASE_MAXFITNESSCOMP ; i++)
		m_fitnessValues[i] = 0;

	m_scaleFactor = 1.0f;
	m_pRootNode = pRootNode;

	rescale();
}

TreeLensInversionGenomeBase::~TreeLensInversionGenomeBase()
{
	delete m_pRootNode;
}
	
bool TreeLensInversionGenomeBase::calculateFitness()
{
	initializeNewCalculation();

	float massum = sumWeights(m_pRootNode, 0, m_pFactory->useWeights());
	float currentBestFitness = 1e20f;
	float currentBestScaleFactor = 1.0f;
	
	float startValue, startValue0;
	float stopValue, stopValue0;
	int numiterations;
	int numiterationsteps;
	
	getCalculationParameters(startValue, stopValue, numiterationsteps, numiterations);

	// we want a scale factor of 1 to correspond to the total mass specified by the
	// massscale used in the BackProjectMatrix

	startValue /= massum;
	stopValue /= massum;

	startValue0 = startValue;
	stopValue0 = stopValue;
	
	for (int i = 0 ; i < numiterations ; i++)
	{
		float stepsize = (stopValue-startValue)/((float)(numiterationsteps-1));
		float s = startValue;

		for (int j = 0 ; j < numiterationsteps ; j++, s += stepsize)
		{
			float f = calculateMassScaleFitness(s);

			if (f < currentBestFitness)
			{
				currentBestFitness = f;
				currentBestScaleFactor = s;
			}
		}

		startValue = currentBestScaleFactor-stepsize;
		stopValue = currentBestScaleFactor+stepsize;

		if (startValue < startValue0)
			startValue = startValue0;
		if (stopValue > stopValue0)
			stopValue = stopValue0;
	}

	m_scaleFactor = currentBestScaleFactor;

	// Do the final (possibly multi-component) fitness evaluation
	calculateTotalFitness(m_scaleFactor, m_fitnessValues);

	return true;
}

mogal::Genome *TreeLensInversionGenomeBase::reproduce(const mogal::Genome *g) const
{
	const TreeLensInversionGenomeBase *pGenome = (const TreeLensInversionGenomeBase *)g;
	
	// Note: using the current reproduction scheme it is not possible that a branch
	//       of the tree gets deeper than the ones of the parents. No check on the
	//       maximum tree depth is necessary.
	GridNodeBase *pRootNode = reproduce(m_pRootNode, pGenome->m_pRootNode, m_scaleFactor, pGenome->m_scaleFactor);

//	std::cout << "----------------------------------------------" << std::endl;
//	std::cout << "PARENT 1: " << std::endl;
//	m_pRootNode->print();
//	std::cout << "PARENT 2: " << std::endl;
//	pGenome->m_pRootNode->print();


	// Make sure we don't exceed the maximum number of basis functions
	
	pRootNode = prune(pRootNode, m_pFactory->getMaximumNumberOfBasisFunctions());
	
//	std::cout << "CHILD: " << std::endl;
//	pRootNode->print();
//	std::cout << "----------------------------------------------" << std::endl;

	return allocateNewInstance(m_pFactory, m_pBackProjectTree, pRootNode);
}

GridNodeBase *TreeLensInversionGenomeBase::prune(GridNodeBase *pNode, int maxBasisFunctions) const
{
	GridNodeBase *pRootNode = pNode;
	std::vector<GridTreeNode *> parents;
	bool done = false;
	SimpleUniformDistribution uniDist(m_pFactory->getRandomNumberGenerator());
	
	while (!done)
	{
		parents.resize(0);
		
		int numberOfWeightNodes = countAllWeightNodes(pRootNode, parents);

		if (numberOfWeightNodes <= maxBasisFunctions)
			done = true;
		else if (parents.empty())
			done = true;
		else
		{
			int idx = (int)((float)parents.size()*(float)uniDist.pickNumber());

			// we'll replace that parent by a random weight node

			GridWeightNode *pWeightNode = new GridWeightNode((float)uniDist.pickNumber(), 0, -1, 0);
			GridNodeBase *pParentsParent = parents[idx]->getParent();
			
			if (pParentsParent == 0) // we're at the root node
			{
				delete pRootNode;
				pRootNode = pWeightNode;

				// after we've replaced the root node by a weight node, there
				// won't be anything else that can be done

				done = true;
			}
			else // the node has a parent itself, attach the new weight node there
			{
				int childIndex = parents[idx]->getChildNumber();

				// this will automatically delete the previous child node

				GridTreeNode *pTreeNode = (GridTreeNode *)pParentsParent;
				
				pTreeNode->setNode(childIndex, pWeightNode);
			}
		}
	}

	return pRootNode;
}

// The parentNodes will contain those tree nodes which only have weight nodes as children
// A check on the level is performed to make sure that the root node is not selected
int TreeLensInversionGenomeBase::countAllWeightNodes(GridNodeBase *pNode, std::vector<GridTreeNode *> &parentNodes)
{
	if (pNode->getNodeType() == GridNodeBase::WeightNode)
		return 1;

	GridTreeNode *pTreeNode = (GridTreeNode *)pNode;
	int numChildren = pTreeNode->getNumberOfChildren();
	int count = 0;

	for (int i = 0 ; i < numChildren ; i++)
		count += countAllWeightNodes(pTreeNode->getChild(i), parentNodes);

	if (count == numChildren && pTreeNode->getParent() != 0) // make sure that we don't select the root node
		parentNodes.push_back(pTreeNode);

	return count;
}

// This preserves as much of the common structure as possible
GridNodeBase *TreeLensInversionGenomeBase::reproduce(const GridNodeBase *pNode1, const GridNodeBase *pNode2, float scaleFactor1, float scaleFactor2) const
{
	GridNodeBase *pNewNode = 0;
	GridNodeBase::GridNodeType nodeType1 = pNode1->getNodeType();
	GridNodeBase::GridNodeType nodeType2 = pNode2->getNodeType();
	SimpleUniformDistribution uniDist(m_pFactory->getRandomNumberGenerator());
	
	if (nodeType1 == nodeType2)
	{
		if (nodeType1 == GridNodeBase::WeightNode) // both are weight nodes, combine them
		{
			// TODO: use the double-peak distribution?

			float x = (float)uniDist.pickNumber();
			const GridWeightNode *pWeightNode1 = (const GridWeightNode *)pNode1;
			const GridWeightNode *pWeightNode2 = (const GridWeightNode *)pNode2;

			if (x < 0.5f)
				pNewNode = new GridWeightNode(pWeightNode1->getWeight()*scaleFactor1, 0, -1, 0);
			else
				pNewNode = new GridWeightNode(pWeightNode1->getWeight()*scaleFactor2, 0, -1, 0);
		}
		else // both are tree nodes, process the child nodes
		{
			const GridTreeNode *pTreeNode1 = (const GridTreeNode *)pNode1;
			const GridTreeNode *pTreeNode2 = (const GridTreeNode *)pNode2;
			int numChildren = pTreeNode1->getNumberOfChildren(); // this should always be the same for the two genomes, only the first level has something different than 4
			GridTreeNode *pNewTreeNode = new GridTreeNode(0, -1, 0, numChildren);
			
			for (int i = 0 ; i < numChildren ; i++)
			{
				GridNodeBase *pNewChildNode = reproduce(pTreeNode1->getChild(i), pTreeNode2->getChild(i), scaleFactor1, scaleFactor2);
				
				pNewTreeNode->setNode(i, pNewChildNode);
			}
			
			pNewNode = pNewTreeNode;
		}
	}
	else // one is a weight node and the other a tree node
	{
		float x = (float)uniDist.pickNumber();

		// std::cout << "Create copy 1" << std::endl;

		if (x < 0.5f)
		{
			pNewNode = pNode1->createCopy();
			scaleAllWeightNodes(pNewNode, scaleFactor1);
		}
		else
		{
			pNewNode = pNode2->createCopy();
			scaleAllWeightNodes(pNewNode, scaleFactor2);
		}
	}

	return pNewNode;
}

void TreeLensInversionGenomeBase::scaleAllWeightNodes(GridNodeBase *pNode, float factor)
{
	if (pNode->getNodeType() == GridNodeBase::WeightNode)
	{
		GridWeightNode *pWeightNode = (GridWeightNode *)pNode;
		float weight = pWeightNode->getWeight();

		pWeightNode->setWeight(weight*factor);
	}
	else // tree node
	{
		GridTreeNode *pTreeNode = (GridTreeNode *)pNode;

		for (int i = 0 ; i < 4 ; i++)
			scaleAllWeightNodes(pTreeNode->getChild(i), factor);
	}
}

mogal::Genome *TreeLensInversionGenomeBase::clone() const
{
	// std::cout << "Creating copy of:" << std::endl;
	// m_pRootNode->print();
	// std::cout << std::endl;

	GridNodeBase *pNewRootNode = m_pRootNode->createCopy();

	// std::cout << "Copy is:" << std::endl;
	// pNewRootNode->print();
	//std::cout << std::endl;

	TreeLensInversionGenomeBase *g = allocateNewInstance(m_pFactory, m_pBackProjectTree, pNewRootNode);
	
	g->setFitnessValues(m_fitnessValues);
	g->setScaleFactor(m_scaleFactor);

	return g;
}

void TreeLensInversionGenomeBase::mutate()
{
//	std::cout << "---------------------------------------------------------------------" << std::endl;
//	std::cout << "BEFORE MUTATION: " << std::endl;
//	m_pRootNode->print();

	std::vector<GridNodeBase *> allNodes;

	findAllWeightNodes(m_pRootNode, allNodes);
	
	// Obtain the number of nodes to mutate (approximate a binomial distribution by a gaussian one)
	
	double p = (double)m_pFactory->getChanceMultiplier()/(double)allNodes.size();

	if (p > 0.99999)
		p = 0.99999;
	
	double mean = p*(double)allNodes.size();
	double sigma = SQRT(mean*(1.0-p));
	SimpleUniformDistribution uniDist(m_pFactory->getRandomNumberGenerator());
	GaussianDistribution gaussDist(mean, sigma, m_pFactory->getRandomNumberGenerator());

	int amountToMutate = (int)(0.5+gaussDist.pickNumber());
	
	if (amountToMutate < 0)
		amountToMutate = 0;
	else if (amountToMutate > allNodes.size())
		amountToMutate = allNodes.size();

//	std::cout << "p = " << p << std::endl;
//	std::cout << "mean = " << mean << std::endl;
//	std::cout << "amountToMutate = " << amountToMutate << std::endl;
		
	for (int i = 0 ; i < amountToMutate ; i++) // we'll choose some to mutate and place them at the front of the array
	{
		int idx = (int)((float)(allNodes.size()-i)*(float)uniDist.pickNumber());

		GridNodeBase *pTmp = allNodes[i];

		allNodes[i] = allNodes[idx+i];
		allNodes[idx+i] = pTmp;
	}

	allNodes.resize(amountToMutate); // ok, the array should now be a lot shorter and
	                                 // should only contain a few nodes to mutate
	
	std::list<GridNodeBase *> nodesToDelete;
	
	for (int i = 0 ; i < amountToMutate ; i++)
	{
		if (allNodes[i]->getLevel() < 0) // this means that we've already removed it, so skip it
		{
			// nothing to do
		}
		else 
		{
			if (allNodes[i]->getNodeType() == GridNodeBase::WeightNode)
			{
//				std::cout << "Mutating a weight node" << std::endl;

				GridWeightNode *pWeightNode = (GridWeightNode *)allNodes[i];
				bool modifyValue = true;

//				std::cout << "Level = " << pWeightNode->getLevel() << " child number = " << pWeightNode->getChildNumber() << std::endl;

				if (pWeightNode->getLevel() < m_pFactory->getMaximumDepth()-1 || pWeightNode->getLevel() > 1) // it's possible to add or remove a layer
				{
					float x = (float)uniDist.pickNumber();

					// TODO: make this configurable?
					if (x > 0.9)
						modifyValue = false;
				}

				if (modifyValue) // just change the value
				{
//					std::cout << "  Changing weight value" << std::endl;
//					std::cout << "    " << pWeightNode->getWeight() << " -> ";

					if (m_pFactory->useAbsoluteMutation())
					{
						pWeightNode->setWeight((float)uniDist.pickNumber());
					}
					else
					{
						float weight = pWeightNode->getWeight();
						float mutationAmplitude = m_pFactory->getMutationAmplitude();
						float p = (float)uniDist.pickNumber()*2.0f-1.0f;
						float x = TAN(p*(float)(CONST_PI/4.0))*mutationAmplitude;
						float y = x+weight;

						if (y < 0)
							y = 0;
						else if (y > 1.0f)
							y = 1.0f;
					
						pWeightNode->setWeight(y);
					}
//					std::cout << pWeightNode->getWeight() << std::endl;
				}
				else // add or remove a layer
				{
					GridTreeNode *pParent = (GridTreeNode *)pWeightNode->getParent();
					int numChildren = pParent->getNumberOfChildren();
					int count = 0;
					bool canRemove = false;
					bool canSpawn = false;
					bool spawn = false;
					
					if (pWeightNode->getLevel() > 1)
					{
						for (int i = 0 ; i < numChildren ; i++)
							if (pParent->getChild(i)->getNodeType() == GridNodeBase::WeightNode)
								count++;

						if (count == numChildren) // it's allowed to remove a small subtree
							canRemove = true;
					}

					if (pWeightNode->getLevel() < m_pFactory->getMaximumDepth()-1)
						canSpawn = true;

					if (canSpawn)
					{
						if (canRemove)
						{
							if ((float)uniDist.pickNumber() < 0.5f)
								spawn = true;
						}
						else
							spawn = true;
					}

					if (spawn)
					{
//						std::cout << "  Spawning subtree" << std::endl;

						GridTreeNode *pTreeNode = new GridTreeNode(0, -1, 0);
						float weight0 = (float)uniDist.pickNumber();
						float weight1 = (float)uniDist.pickNumber();
						float weight2 = (float)uniDist.pickNumber();
						float weight3 = (float)uniDist.pickNumber();

						float sum = weight0+weight1+weight2+weight3;
						float oldWeight = pWeightNode->getWeight();

						// TODO: this should probably be modified if cell weights are being
						// used
		
						GridWeightNode *pChild0 = new GridWeightNode(oldWeight*weight0/sum, 0, -1, 0);
						GridWeightNode *pChild1 = new GridWeightNode(oldWeight*weight1/sum, 0, -1, 0);
						GridWeightNode *pChild2 = new GridWeightNode(oldWeight*weight2/sum, 0, -1, 0);
						GridWeightNode *pChild3 = new GridWeightNode(oldWeight*weight3/sum, 0, -1, 0);

						pTreeNode->setNode(0, pChild0);
						pTreeNode->setNode(1, pChild1);
						pTreeNode->setNode(2, pChild2);
						pTreeNode->setNode(3, pChild3);

						// Okay, we've build the subtree, now extract the old weight node
						// and attach this one instead.
						// Note that the weight node is never the root node
		
						GridNodeBase *pExtractedNode = 0;
						int childNumber = pWeightNode->getChildNumber();

						pExtractedNode = pParent->extractChild(childNumber);
						pParent->setNode(childNumber, pTreeNode);

						pExtractedNode->setLevel(-10*m_pFactory->getMaximumDepth());
						nodesToDelete.push_back(pExtractedNode);
					}
					else // remove a subtree
					{
//						std::cout << "  Removing a subtree" << std::endl;
						
						// Here, we should remove the parent itself and replace
						// it with a weight node. Again, the parent should never
						// be able to be the root node

						GridTreeNode *pParentsParent = (GridTreeNode *)pParent->getParent();
						int index = pParent->getChildNumber();
						GridWeightNode *pNewWeight = new GridWeightNode((float)uniDist.pickNumber(), 0, -1, 0);

						GridNodeBase *pExtractedNode = pParentsParent->extractChild(index);
						pParentsParent->setNode(index, pNewWeight);

						pExtractedNode->setLevel(-10*m_pFactory->getMaximumDepth());
						nodesToDelete.push_back(pExtractedNode);
					}
				}
			}
			else
			{
// 				std::cout << "ERROR" << std::endl;
			}
		}
	}

	while (!nodesToDelete.empty())
	{
		delete *(nodesToDelete.begin());
		nodesToDelete.pop_front();
	}

//	std::cout << std::endl << "AFTER MUTATION: " << std::endl;
//	m_pRootNode->print();

	// make sure that we don't exceed the maximum number of basis functions
	m_pRootNode = prune(m_pRootNode, m_pFactory->getMaximumNumberOfBasisFunctions());

//	std::cout << std::endl << "AFTER PRUNE: " << std::endl;
//	m_pRootNode->print();

//	std::cout << "---------------------------------------------------------------------" << std::endl;

	rescale();
}

void TreeLensInversionGenomeBase::findAllWeightNodes(GridNodeBase *pNode, std::vector<GridNodeBase *> &allNodes)
{
	if (pNode->getNodeType() == GridNodeBase::TreeNode)
	{
		GridTreeNode *pTreeNode = (GridTreeNode *)pNode;
		int numChildren = pTreeNode->getNumberOfChildren();

		for (int i = 0 ; i < numChildren ; i++)
			findAllWeightNodes(pTreeNode->getChild(i), allNodes);
	}
	else
		allNodes.push_back(pNode);
}

std::string TreeLensInversionGenomeBase::getFitnessDescription() const
{
	char str[1024];
	char *p = str;
	int num = getFactory()->getNumberOfFitnessComponents();

	for (int i = 0 ; i < num ; i++)
	{
		sprintf(p,"%g ",(double)m_fitnessValues[i]);
		p += strlen(p);
	}
	return std::string(str);
}
	
void TreeLensInversionGenomeBase::rescale()
{
	float maxWeight = findMaxWeight(m_pRootNode);
	
	scaleWeights(m_pRootNode, 0.5f/maxWeight);
}

float TreeLensInversionGenomeBase::findMaxWeight(const GridNodeBase *pNode)
{
	float maxWeight = 0;
	
	if (pNode->getNodeType() == GridNodeBase::TreeNode)
	{
		const GridTreeNode *pTreeNode = (const GridTreeNode *)pNode;
		int numChildren = pTreeNode->getNumberOfChildren();

		maxWeight = findMaxWeight(pTreeNode->getChild(0));
		
		for (int i = 1 ; i < numChildren ; i++)
			maxWeight = MAX(maxWeight, findMaxWeight(pTreeNode->getChild(i)));
	}
	else // weight node
	{
		const GridWeightNode *pWeightNode = (const GridWeightNode *)pNode;
		
		maxWeight = pWeightNode->getWeight();
	}

	return maxWeight;
}

float TreeLensInversionGenomeBase::sumWeights(const GridNodeBase *pNode, int level, bool useWeights)
{
	float sum = 0;
	
	if (pNode->getNodeType() == GridNodeBase::TreeNode)
	{
		const GridTreeNode *pTreeNode = (const GridTreeNode *)pNode;
		int numChildren = pTreeNode->getNumberOfChildren();

		sum = sumWeights(pTreeNode->getChild(0), level+1, useWeights);

		for (int i = 0 ; i < numChildren ; i++)
			sum += sumWeights(pTreeNode->getChild(i), level+1, useWeights);
	}
	else // weight node
	{
		const GridWeightNode *pWeightNode = (const GridWeightNode *)pNode;
		float factor2 = 1.0f;

		if (useWeights)
		{
			int factor = 1 << level;
			factor2 = (float)factor*(float)factor;
		}	
		
		sum = pWeightNode->getWeight()/factor2;
	}

	return sum;
}

void TreeLensInversionGenomeBase::scaleWeights(GridNodeBase *pNode, float scaleFactor)
{
	if (pNode->getNodeType() == GridNodeBase::TreeNode)
	{
		GridTreeNode *pTreeNode = (GridTreeNode *)pNode;
		int numChildren = pTreeNode->getNumberOfChildren();

		for (int i = 0 ; i < numChildren ; i++)
			scaleWeights(pTreeNode->getChild(i), scaleFactor);
	}
	else // weight node
	{
		GridWeightNode *pWeightNode = (GridWeightNode *)pNode;
		float oldWeight = pWeightNode->getWeight();
		
		pWeightNode->setWeight(oldWeight*scaleFactor);
	}
}

GravitationalLens *TreeLensInversionGenomeBase::createLens(double *totalmass, std::string &errorString) const
{
	const TreeLensInversionGAFactoryParams *pParams = (const TreeLensInversionGAFactoryParams *)m_pFactory->getCurrentParameters();
	double massScale = pParams->getMassScale();
	double tm = 0;
	GravitationalLens *pLens = 0;
	Vector2D<double> center = pParams->getCenter();
	double areaSize = pParams->getAreaSize();
	double widthScaleFactor = pParams->getWidthScaleFactor();
	
	if (pParams->getBasisFunctionType() == TreeLensInversionGAFactoryParams::PlummerBasis)
	{
		MultiplePlummerLensParams lensParams;
		pLens = new MultiplePlummerLens();

		addBasisFunction(m_pRootNode, 0, &lensParams, center, areaSize, widthScaleFactor, massScale);
		
		if (!pLens->init(pParams->getD_d(), &lensParams))
		{
			errorString = pLens->getErrorString();
			delete pLens;
			return 0;
		}
	}
	else // unsupported basis type
	{
		errorString = std::string("Unsupported basis function type");
		return 0;
	}
	*totalmass = tm;

	return pLens;
}

void TreeLensInversionGenomeBase::addBasisFunction(const GridNodeBase *pNode, int level, MultiplePlummerLensParams *pLensParams, Vector2D<double> center, double areaSize, double widthScaleFactor, double massScale) const
{
	if (pNode->getNodeType() == GridNodeBase::TreeNode)
	{
		const GridTreeNode *pTreeNode = (const GridTreeNode *)pNode;
		int numChildren = pTreeNode->getNumberOfChildren();
		double numSubdivDouble = SQRT((double)numChildren);
		int numSubdiv = (int)numSubdivDouble;

		if (numSubdiv*numSubdiv < numChildren)
			numSubdiv++;
		
		double subAreaSize = areaSize/(double)numSubdiv;

		for (int y = 0 ; y < numSubdiv ; y++)
		{
			for (int x = 0 ; x < numSubdiv ; x++)
			{
				double xPos = subAreaSize/2.0 + center.getX() - areaSize/2.0 + ((double)x)*subAreaSize;
				double yPos = subAreaSize/2.0 + center.getY() - areaSize/2.0 + ((double)y)*subAreaSize;

				addBasisFunction(pTreeNode->getChild((y*numSubdiv)+x), level+1, pLensParams, Vector2D<double>(xPos, yPos), subAreaSize, widthScaleFactor, massScale);
			}
		}
	}
	else // weight node
	{
		const GridWeightNode *pWeightNode = (const GridWeightNode *)pNode;
		double weight = (double)pWeightNode->getWeight();
		float factor2 = 1.0f;

		if (m_pFactory->useWeights())
		{
			int factor = 1<<level;
			
			factor2 = (float)factor*(float)factor;
		}
		
		// TODO: remove plummer specific references?
		pLensParams->addLensInfo(PlummerLensInfo(m_scaleFactor*massScale*weight/factor2, areaSize*widthScaleFactor, center));
	}
}

} // end namespace

