/*

  This file is a part of GRALE, a library to facilitate the simulation
  and inversion of gravitational lenses.

  Copyright (C) 2008-2012 Jori Liesenborgs

  Contact: jori.liesenborgs@gmail.com
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
*/

#ifndef GRALE_IMAGESDATA_H

#define GRALE_IMAGESDATA_H

#include "graleconfig.h"
#include "vector2d.h"
#include "triangle2d.h"
#include "triangulation.h"
#include <serut/serializationinterface.h>
#include <sys/types.h>
#include <stdint.h>
#include <string>
#include <vector>
#include <list>

namespace grale
{

class CoordinateInterface;	

template<class T> class Polygon2D;

class GRALE_IMPORTEXPORT ImagesData : public errut::ErrorBase
{
public:
	ImagesData();
	ImagesData(const ImagesData &data)						{ copyFrom(data); }
	virtual ~ImagesData();
	ImagesData *createCopy() const;

	bool createGridImage(Vector2D<double> bottomleft, Vector2D<double> topright, int numx, int numy, 
	                     const std::list<Triangle2D<double> > &excluderegions, bool triangulate = true,
			     bool useIntens = false);
	bool createGridImage(Vector2D<double> bottomleft, Vector2D<double> topright, int numx, int numy, 
	                     const std::list<ImagesData *> &excludeImages, bool useIntens = false,
			     bool storeMinimalDistances = false);
	
	bool create(int numImages, bool intensities, bool shearInfo);

	bool addGridImage(Vector2D<double> bottomleft, Vector2D<double> topright, int numx, int numy, 
	                  const std::list<Triangle2D<double> > &excluderegions, bool triangulate = true);

	int addImage();
	int addPoint(int imageNumber, Vector2D<double> point); // returns index of the point, -1 on error
	int addPoint(int imageNumber, Vector2D<double> point, double intensity); // returns index of the point, -1 on error
	int addPoint(int imageNumber, Vector2D<double> point, double shearComponent1, double shearComponent2);
	int addPoint(int imageNumber, Vector2D<double> point, double intensity, double shearComponent1, double shearComponent2);
	//bool setImagePointIntensity(int imageNumber, int pointNumber, double intensity);

	int addGroup(); // returns the index of the group, -1 on error
	bool addGroupPoint(int groupNumber, int imageIndex, int pointIndex);
	bool addTimeDelayInfo(int imageIndex, int pointIndex, double timeDelay); // timeDelay in days
	bool addTriangle(int imageNumber, int index1, int index2, int index3);
	
	bool load(const std::string &fname);
	bool save(const std::string &fname) const;
	bool read(serut::SerializationInterface &si);
	bool write(serut::SerializationInterface &si) const;
	bool loadPNG(const std::string &fname, Vector2D<double> bottomleft, Vector2D<double> topright,
	             bool nullspace = false, const std::string &maskfile = std::string(""), bool triangulate = true);
	bool loadFITS(const std::string &fname, bool nullspace = false, 
	              const std::string &maskfile = std::string(""), bool triangulate = true);
	bool loadFITS(const std::string &fname, double unit, bool nullspace = false, 
	              const std::string &maskfile = std::string(""), bool triangulate = true);
	bool loadFITS(const std::string &fname, Vector2D<double> bottomleft, Vector2D<double> topright,
	              bool nullspace = false, const std::string &maskfile = std::string(""), bool triangulate = true);

	bool hasIntensities() const							{ if (m_intensities.empty()) return false; return true; }
	bool hasShearInfo() const							{ if (m_shearComponent1s.empty()) return false; return true; }
	bool hasTimeDelays() const							{ if (m_timeDelayInfo.empty()) return false; return true; }

	int getNumberOfImages() const 							{ return m_images.size(); }
	int getNumberOfImagePoints(int i) const						{ return m_images[i].size(); }
	Vector2D<double> getImagePointPosition(int image, int point) const		{ return m_images[image][point]; }
	double getImagePointIntensity(int image, int point) const			{ return m_intensities[image][point]; }
	double getShearComponent1(int image, int point) const				{ return m_shearComponent1s[image][point]; }
	double getShearComponent2(int image, int point) const				{ return m_shearComponent2s[image][point]; }

	bool createTriangulation();
	bool clearTriangulation();
	bool hasTriangulation()	const							{ if (m_triangulations.size() != 0) return true; return false; }
	bool getTriangles(int image, std::list<TriangleIndices> &triangles) const;
	bool getImageBorder(int image, Polygon2D<double> &polygon) const;

	int getNumberOfGroups() const							{ return m_groupPoints.size(); }
	int getNumberOfGroupPoints(int group) const					{ return m_groupPoints[group].size()/2; }
	void getGroupPointIndices(int group, int pointnr, int *img, int *point) const	{ *img = m_groupPoints[group][pointnr*2]; *point = m_groupPoints[group][pointnr*2+1]; }
	void clearGroupInfo()								{ m_groupPoints.clear(); }

	int getNumberOfTimeDelays() const						{ return m_timeDelayInfo.size(); }
	void getTimeDelay(int index, int *pImg, int *pPoint, double *pDelay) const	{ *pImg = m_timeDelayInfo[index].getImageIndex(); *pPoint = m_timeDelayInfo[index].getPointIndex(); *pDelay = m_timeDelayInfo[index].getTimeDelay(); }
	void clearTimeDelayInfo() 							{ m_timeDelayInfo.clear(); }

	Vector2D<double> getTopRightCorner() const					{ return m_topRight; }
	Vector2D<double> getBottomLeftCorner() const					{ return m_bottomLeft; }

	void subtractPosition(Vector2D<double> pos);
	void subtractIntensity(double v);

	const ImagesData &operator=(const ImagesData &dat)				{ clear(); copyFrom(dat); return *this; }
private:
	void clear();
	void findExtremes();
	void copyFrom(const ImagesData &data);
	bool processData(double *data, CoordinateInterface &coordcalc, int numx, int numy, bool intens, bool nullspace, const std::string &maskfile, bool triangulate);
	bool createMap(double *values, int numx, int numy, const std::string &maskfile, std::vector<uint16_t> &map);
	static int processTrianglePoint(Vector2D<double> p, std::vector<Vector2D<double> > &points);
	static int processTrianglePoint(Vector2D<double> p, std::vector<Vector2D<double> > &points, std::vector<double> &intensities);
	static void createLineSegments(const std::list<IntVector2DPlus> &points, std::list<IntLine2D> &lines);
	static void addLineSegment(std::list<IntLine2D> &lines, IntLine2D l);
	static void createTriangulation(const std::list<IntLine2D> &lines, std::list<TriangleIndices> &triangles, int numx);

	class TimeDelayPoint
	{
	public:
		TimeDelayPoint(int imageIndex, int pointIndex, double timeDelay)
		{
			m_imageIndex = imageIndex;
			m_pointIndex = pointIndex;
			m_timeDelay = timeDelay;
		}

		int getImageIndex() const						{ return m_imageIndex; }
		int getPointIndex() const						{ return m_pointIndex; }
		double getTimeDelay() const						{ return m_timeDelay; }
	private:
		int m_imageIndex;
		int m_pointIndex;
		double m_timeDelay;
	};

	std::vector<std::vector<Vector2D<double> > > m_images;
	std::vector<std::vector<double> > m_intensities;
	std::vector<std::vector<double> > m_shearComponent1s;
	std::vector<std::vector<double> > m_shearComponent2s;
	std::vector<std::vector<int32_t> > m_groupPoints;
	std::vector<std::list<TriangleIndices> > m_triangulations;
	std::vector<TimeDelayPoint> m_timeDelayInfo;
	Vector2D<double> m_bottomLeft;
	Vector2D<double> m_topRight;
};

class ImagesDataExtended : public ImagesData
{
public:
	ImagesDataExtended()							{ D_s = 0; D_ds = 0; }
	ImagesDataExtended(const ImagesDataExtended &dat) : ImagesData(dat) 	{ D_s = dat.D_s; D_ds = dat.D_ds; std::list<double>::const_iterator it; for (it = dat.extraparameters.begin() ; it != dat.extraparameters.end() ; it++) extraparameters.push_back(*it); }
	ImagesDataExtended(double D_s,double D_ds)				{ ImagesDataExtended::D_s = D_s; ImagesDataExtended::D_ds = D_ds; }
	~ImagesDataExtended()							{ }
	double getD_s()	const							{ return D_s; }
	double getD_ds() const							{ return D_ds; }
	void setD_s(double v)							{ D_s = v; }
	void setD_ds(double v)							{ D_ds = v; }

	bool read(serut::SerializationInterface &si);
	bool write(serut::SerializationInterface &si) const;

	const ImagesDataExtended &operator=(const ImagesDataExtended &dat)	 { D_s = dat.D_s; D_ds = dat.D_ds; ImagesData::operator=(dat); return *this; }

	void clearExtraParameters()						{ extraparameters.clear(); }
	void addExtraParameter(double val)					{ extraparameters.push_back(val); }
	int getNumberOfExtraParameters() const					{ return extraparameters.size(); }
	bool getExtraParameter(int idx, double *val) const			{ if (idx < 0 || idx >= extraparameters.size()) { setErrorString("Index out of bounds"); return false; } std::list<double>::const_iterator it = extraparameters.begin(); while (idx > 0) { it++; idx--; } *val = *it; return true; }
private:
	double D_s,D_ds;
	std::list<double> extraparameters;
};

} // end namespace

#endif // GRALE_IMAGESDATA_H

