/*

  This file is a part of GRALE, a library to facilitate the simulation
  and inversion of gravitational lenses.

  Copyright (C) 2008-2012 Jori Liesenborgs

  Contact: jori.liesenborgs@gmail.com
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
*/

#include "testlens.h"
#include "circularlensprofile.h"
#include "constants.h"

namespace grale
{

class TestNSISProfile : public CircularLensProfile
{
public:
	TestNSISProfile(GravitationalLens *pLens, double sigmav, double thetac)
	{
		m_pLens = pLens;
		m_sigmav2 = sigmav*sigmav;
		m_thetac = thetac;
		m_thetac2 = thetac*thetac;
	}

	double getMassInside(double theta)
	{
		return CONST_PI*m_sigmav2*m_pLens->getLensDistance()/CONST_G*(SQRT(theta*theta+m_thetac2)-m_thetac);
	}

	double getSurfaceMassDensity(double theta)
	{
		return m_sigmav2/(2.0*CONST_G*m_pLens->getLensDistance()*SQRT(theta*theta+m_thetac2));
	}

	double getSurfaceMassDensityDerivativeOverTheta(double theta)
	{
		double part = theta*theta+m_thetac2;
		return -m_sigmav2/(2.0*CONST_G*m_pLens->getLensDistance()*SQRT(part)*part);
	}
private:
	GravitationalLens *m_pLens;
	double m_sigmav2;
	double m_thetac, m_thetac2;
};

class TestSISProfile : public CircularLensProfile
{
public:
	TestSISProfile(GravitationalLens *pLens, double sigmav)
	{
		m_pLens = pLens;
		m_sigmav2 = sigmav*sigmav;
	}

	double getMassInside(double theta)
	{
		return CONST_PI*m_sigmav2*m_pLens->getLensDistance()/CONST_G*theta;
	}

	double getSurfaceMassDensity(double theta)
	{
		return m_sigmav2/(2.0*CONST_G*m_pLens->getLensDistance()*theta);
	}

	double getSurfaceMassDensityDerivativeOverTheta(double theta)
	{
		return -m_sigmav2/(2.0*CONST_G*m_pLens->getLensDistance()*theta*theta*theta);
	}
private:
	GravitationalLens *m_pLens;
	double m_sigmav2;
};

GravitationalLensParams *TestLensParams::createCopy() const
{
	return new TestLensParams();
}

TestLens::TestLens() : EllipticLens(TestLensType)
{
}

TestLens::~TestLens()
{
	delete m_pProfile;
}

bool TestLens::processParameters(const GravitationalLensParams *pLensParams)
{
	m_q = 0.9;
//	m_q = 1.0;
	//m_pProfile = new TestNSISProfile(this, 1200000.0, 10.0*ANGLE_ARCSEC);
	//m_pProfile = new TestNSISProfile(this, 1200000.0, 0.1*ANGLE_ARCSEC);
	m_pProfile = new TestSISProfile(this, 1200000.0);

	subInit(m_q, m_pProfile, 0, 1e-5, 1024);
	return true;
}

} // end namespace
