/*

  This file is a part of GRALE, a library to facilitate the simulation
  and inversion of gravitational lenses.

  Copyright (C) 2008-2012 Jori Liesenborgs

  Contact: jori.liesenborgs@gmail.com
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
*/

#include "graleconfig.h"
#include "circularintegrator.h"
#include "constants.h"
#include <vector>

#include "debugnew.h"

namespace grale
{

CircularIntegrator::CircularIntegrator(Vector2D<double> center, double startRadius, double stopRadius, 
		                       int radialSteps, int angleSteps)
{
	m_center = center;
	m_startRadius = ABS(startRadius);
	m_stopRadius = ABS(stopRadius);

	if (m_startRadius > m_stopRadius)
	{
		m_startRadius = ABS(stopRadius);
		m_stopRadius = ABS(startRadius);
	}
	
	m_rSteps = radialSteps;
	m_aSteps = angleSteps;

	if (m_aSteps < 1)
		m_aSteps = 1;
	if (m_rSteps < 1)
		m_rSteps = 1;

	m_angularStepSize = CONST_PI*2.0/(double)m_aSteps;
	m_radialStepSize = (m_stopRadius-m_startRadius)/(double)m_rSteps;
}

CircularIntegrator::~CircularIntegrator()
{
}

double CircularIntegrator::integrate(Real2DFunction &f, bool abs) const
{
	double total = 0;
	double r;
	int i, j;
	
	std::vector<std::vector<double> > heights;

	heights.resize(m_rSteps+1);
	for (j = 0 ; j <= m_rSteps ; j++)
		heights[j].resize(m_aSteps);

	for (i = 0, r = m_startRadius ; i <= m_rSteps ; i++, r += m_radialStepSize)
	{
		double a;

		for (j = 0, a = 0 ; j < m_aSteps ; j++, a += m_angularStepSize)
		{
			Vector2D<double> offset(r*COS(a), r*SIN(a));
			Vector2D<double> point = m_center + offset;

			double h = f(point);
			
			if (abs)
				h = ABS(h);

			heights[i][j] = h;
		}
	}
	
	for (i = 0, r = m_startRadius ; i < m_rSteps ; i++, r += m_radialStepSize)
	{
		double rNext = r + m_radialStepSize;

		for (j = 0 ; j < m_aSteps ; j++)
		{
			double h1 = heights[i][j];
			double h2 = heights[i+1][j];
			
			int jNext = (j+1)%m_aSteps;

			double h3 = heights[i][jNext];
			double h4 = heights[i+1][jNext];

			double p1 = h1+h3;
			double p2 = h2+h4;

			total += (1.0/12.0)*m_radialStepSize*m_angularStepSize*(r*(2.0*p1+p2)+rNext*(p1+2.0*p2));
		}
	}
	
	return total;
}

} // end namespace

