/*

  This file is a part of GRALE, a library to facilitate the simulation
  and inversion of gravitational lenses.

  Copyright (C) 2008-2012 Jori Liesenborgs

  Contact: jori.liesenborgs@gmail.com
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
*/

#include "graleconfig.h"
#include "gridfunction.h"

#include <iostream>

#include "debugnew.h"

namespace grale
{

GridFunction::GridFunction(Real2DFunction &f, Vector2D<double> bottomLeft, Vector2D<double> topRight,
                           int numX, int numY, bool abs)
{
	if (numX < 10)
		m_numX = 10;
	else
		m_numX = numX;
	if (numY < 10)
		m_numY = 10;
	else
		m_numY = numY;

	int numTotal = m_numX*m_numY;
	m_pValues = new double[numTotal];

	m_xStep = (topRight.getX() - bottomLeft.getX())/((double)(m_numX-1));
	m_yStep = (topRight.getY() - bottomLeft.getY())/((double)(m_numY-1));
	m_topRight = topRight;
	m_bottomLeft = bottomLeft;
	
	double x2, y2;
	int x, y;
	int i = 0;

	if (!abs)
	{
		for (y = 0, y2 = m_bottomLeft.getY() ; y < m_numY ; y++, y2 += m_yStep)
		{
			for (x = 0, x2 = m_bottomLeft.getX() ; x < m_numX ; x++, x2 += m_xStep)
			{
				m_pValues[i] = f(Vector2D<double>(x2, y2));
				i++;
			}
		}
	}
	else
	{
		for (y = 0, y2 = m_bottomLeft.getY() ; y < m_numY ; y++, y2 += m_yStep)
		{
			for (x = 0, x2 = m_bottomLeft.getX() ; x < m_numX ; x++, x2 += m_xStep)
			{
				m_pValues[i] = ABS(f(Vector2D<double>(x2, y2)));
				i++;
			}
		}	
	}

	m_deleteValues = true;
}

GridFunction::GridFunction(const double *pValues, Vector2D<double> bottomLeft, Vector2D<double> topRight,
		           int numX, int numY, bool deleteValues)
{
	m_numX = numX;
	m_numY = numY;

	int numTotal = m_numX*m_numY;
	//m_pValues = new double[numTotal];
	m_pValues = (double *)pValues;

	m_xStep = (topRight.getX() - bottomLeft.getX())/((double)(m_numX-1));
	m_yStep = (topRight.getY() - bottomLeft.getY())/((double)(m_numY-1));
	m_topRight = topRight;
	m_bottomLeft = bottomLeft;
	
	//for (int i = 0 ; i < numTotal ; i++)
	//	m_pValues[i] = pValues[i];

	m_deleteValues = deleteValues;
}

GridFunction::GridFunction(const std::list<Triangle2DPlus<double> > &triangles ,Vector2D<double> bottomLeft, Vector2D<double> topRight, 
	             int numX, int numY, bool abs, bool fast, int resampleSize)
{
	if (resampleSize < 1)
		resampleSize = 1;
	
	m_numX = numX*resampleSize;
	m_numY = numY*resampleSize;

	int numTotal = m_numX*m_numY;
	m_pValues = new double[numTotal];

	m_xStep = (topRight.getX() - bottomLeft.getX())/((double)(m_numX-1));
	m_yStep = (topRight.getY() - bottomLeft.getY())/((double)(m_numY-1));
	m_topRight = topRight;
	m_bottomLeft = bottomLeft;

	if (!fast)
	{
		double x2, y2;
		int x, y;
		int i = 0;

		for (y = 0, y2 = m_bottomLeft.getY() ; y < m_numY ; y++, y2 += m_yStep)
		{
			for (x = 0, x2 = m_bottomLeft.getX() ; x < m_numX ; x++, x2 += m_xStep, i++)
			{
				std::list<Triangle2DPlus<double> >::const_iterator it;
				double totalHeight = 0;
				int numHits = 0;

				for (it = triangles.begin() ; it != triangles.end() ; it++)
				{
					Triangle2DPlus<double> triangle = (*it);
					const Vector2D<double> *pPoints = triangle.getPoints();
					double minX = pPoints[0].getX();
					double maxX = minX;
					double minY = pPoints[0].getY();
					double maxY = minY;

					for (int k = 1 ; k < 3 ; k++)
					{
						minX = MIN(minX, pPoints[k].getX());
						maxX = MAX(maxX, pPoints[k].getX());
						minY = MIN(minY, pPoints[k].getY());
						maxY = MAX(maxY, pPoints[k].getY());
					}

					if (!(x2 < minX || x2 > maxX || y2 < minY || y2 > maxY))
					{
						// intersection possible

						double height;

						if (triangle.isInside(Vector2D<double>(x2,y2), height))
						{
							totalHeight += height;
							numHits++;
						}
					}
				}

				if (numHits > 0)
					totalHeight /= (double)numHits;

				m_pValues[i] = totalHeight;
			}
		}
	}
	else
	{
		int *pPixelCount = new int[numTotal];

		for (int i = 0 ; i < numTotal ; i++)
		{
			pPixelCount[i] = 0;
			m_pValues[i] = 0;
		}
		std::list<Triangle2DPlus<double> >::const_iterator it;

		for (it = triangles.begin() ; it != triangles.end() ; it++)
			drawTriangle(*it, pPixelCount);

		for (int i = 0 ; i < numTotal ; i++)
		{
			int div = MAX(1, pPixelCount[i]);
			m_pValues[i] /= (double)div;
		}
		delete [] pPixelCount;
	}

	// Now, resample
	
	if (resampleSize > 1)
	{
		double resampleDiv = (double)(resampleSize*resampleSize);
		int largeNumX = m_numX;
		int largeNumY = m_numY;
		
		m_numX = numX;
		m_numY = numY;
		
		int numTotal = m_numX*m_numY;
		double *pNewValues = new double [numTotal];

		int idx = 0;
		int pos = 0;
		int largeLineStep = largeNumX - resampleSize;
		for (int y = 0 ; y < m_numY ; y++)
		{
			for (int x = 0 ; x < m_numX ; x++, idx++)
			{
				double sum = 0;
				int pos = (x*resampleSize)+(y*resampleSize)*largeNumX;
				
				for (int j = 0 ; j < resampleSize ; j++)
				{
					for (int i = 0 ; i < resampleSize ; i++, pos++)
						sum += m_pValues[pos];
					pos += largeLineStep;
				}
				
				pNewValues[idx] = sum/resampleDiv;
			}
		}

		double *pTmp = m_pValues;
		m_pValues = pNewValues;
		delete [] pTmp;
		
		m_xStep = (topRight.getX() - bottomLeft.getX())/((double)(m_numX-1));
		m_yStep = (topRight.getY() - bottomLeft.getY())/((double)(m_numY-1));
	}

	m_deleteValues = true;
}

GridFunction::~GridFunction()
{
	if (m_deleteValues)
		delete [] m_pValues;
}

double GridFunction::operator()(Vector2D<double> v) const
{
	double xDist = v.getX() - m_bottomLeft.getX();
	double yDist = v.getY() - m_bottomLeft.getY();
	
	double xFracDist = xDist/m_xStep;
	double yFracDist = yDist/m_yStep;
	
	int x = (int)(xFracDist);
	int y = (int)(yFracDist);

	double t = xFracDist-(double)x;
	double u = yFracDist-(double)y;

	if (x < 0 || y < 0 || x >= (m_numX-1) || y >= (m_numY-1))
		return 0;
	
	double y1 = m_pValues[x + m_numX*y];
	double y2 = m_pValues[x + 1 + m_numX*y];
	double y3 = m_pValues[x + 1 + m_numX*(y + 1)];
	double y4 = m_pValues[x + m_numX*(y + 1)];

	double val = (1.0-t)*(1.0-u)*y1 + t*(1.0-u)*y2 + t*u*y3 + (1.0-t)*u*y4;
	return val;
}

Vector2D<double> GridFunction::getGradient(Vector2D<double> v) const
{
	double a, x, xDiff;
	double y, yDiff;

	xDiff = m_xStep/10.0;
	yDiff = m_yStep/10.0;

	a = operator()(v);
	x = operator()(v + Vector2D<double>(xDiff, 0.0));
	y = operator()(v + Vector2D<double>(0.0, yDiff));
	return Vector2D<double>((x-a)/xDiff, (y-a)/yDiff);
}

void GridFunction::drawTriangle(const Triangle2DPlus<double> &triangle, int *pPixelCount)
{
	const Vector2D<double> *pPoints = triangle.getPoints();
	const double *pHeights = triangle.getHeights();

	// sort the Y coordinates
	
	double sortedX[3];
	double sortedY[3];
	double sortedZ[3];

	int order[3] = { -1, -1, -1 };

	if (pPoints[0].getY() < pPoints[1].getY())
	{
		if (pPoints[2].getY() < pPoints[0].getY())
		{
			order[0] = 2;
			order[1] = 0;
			order[2] = 1;
		}
		else if (pPoints[2].getY() >= pPoints[1].getY())
		{
			order[0] = 0;
			order[1] = 1;
			order[2] = 2;
		}
		else
		{
			order[0] = 0;
			order[1] = 2;
			order[2] = 1;
		}
	}
	else
	{
		if (pPoints[2].getY() < pPoints[1].getY())
		{
			order[0] = 2;
			order[1] = 1;
			order[2] = 0;
		}
		else if (pPoints[2].getY() >= pPoints[0].getY())
		{
			order[0] = 1;
			order[1] = 0;
			order[2] = 2;
		}
		else
		{
			order[0] = 1;
			order[1] = 2;
			order[2] = 0;
		}
	}

	for (int i = 0 ; i < 3 ; i++)
	{
		int idx = order[i];
		sortedX[i] = pPoints[idx].getX();
		sortedY[i] = pPoints[idx].getY();
		sortedZ[i] = pHeights[idx];
	}

	int intX[3];
	int intY[3];

	for (int i = 0 ; i < 3 ; i++)
	{
		intX[i] = (int)((sortedX[i]-m_bottomLeft.getX())/m_xStep+0.5);
		intY[i] = (int)((sortedY[i]-m_bottomLeft.getY())/m_yStep+0.5);
	}

	drawIntTriangle(intX, intY, sortedZ, pPixelCount);
}

void GridFunction::drawIntTriangle(const int *pX, const int *pY, const double *pZ, int *pPixelCount)
{
	std::vector<PixelValue> line[3];

	processLine(pX[0], pY[0], pZ[0], pX[1], pY[1], pZ[1], line[0]);
	processLine(pX[0], pY[0], pZ[0], pX[2], pY[2], pZ[2], line[1]);
	processLine(pX[1], pY[1], pZ[1], pX[2], pY[2], pZ[2], line[2]);

	int baseY = pY[0];
	int numLines = pY[2] - pY[0] + 1;
	std::vector<MinMaxValue> lineInfo(numLines);

	for (int j = 0 ; j < 3 ; j++)
	{
		for (int i = 0 ; i < line[j].size() ; i++)
		{
			int idx = line[j][i].getY() - baseY;
			
			lineInfo[idx].process(line[j][i].getX(), line[j][i].getValue());
		}
	}

	int y = baseY;
	
	for (int i = 0 ; i < numLines ; i++)
	{
		if (!(y < 0 || y > (m_numY-1)))
		{
			// TODO: for debugging!
			if (!lineInfo[i].isSet())
				std::cerr << "WARNING: index " << i << " is not set!" << std::endl;

			int minX = lineInfo[i].getMinX();
			int maxX = lineInfo[i].getMaxX();
			double startValue = lineInfo[i].getMinValue();
			double stopValue = lineInfo[i].getMaxValue();

			int steps = maxX - minX + 1;
			int x = minX;
			double dh;
			double h = startValue;
			
			if (steps > 1)
				dh = (stopValue - startValue)/((double)(steps-1));
			else
				dh = 0;

			for (int j = 0 ; j < steps-1 ; j++)
			{
				if (!(x < 0 || x > (m_numX-1)))
				{
					int pos = x+y*m_numX;
					m_pValues[pos] += h;
					pPixelCount[pos]++;
				}
				x++;
				h += dh;
			}

			x = maxX;
			h = stopValue;

			if (!(x < 0 || x > (m_numX-1)))
			{
				int pos = x+y*m_numX;
				m_pValues[pos] += h;
				pPixelCount[pos]++;
			}
		}
		y++;
	}
}

void GridFunction::processLine(int x1, int y1, double z1, int x2, int y2, double z2, std::vector<PixelValue> &pixelValues)
{
	if (y2 == y1)
	{
		pixelValues.resize(2);
		pixelValues[0] = PixelValue(x1, y1, z1);
		pixelValues[1] = PixelValue(x2, y2, z2);
		return;
	}
	
	int steps = (y2-y1)+1;
	double dx = ((double)(x2 - x1))/((double)(steps - 1));
	double dh = ((double)(z2 - z1))/((double)(steps - 1));
	
	pixelValues.resize(steps);
	int y = y1;
	double x = x1;
	double h = z1;

	for (int i = 0 ; i < steps - 1 ; i++)
	{
		pixelValues[i] = PixelValue((int)(x + 0.5), y, h);

		h += dh;
		x += dx;
		y++;
	}

	x = x2;
	h = z2;
	y = y2;
	pixelValues[steps - 1] = PixelValue((int)(x + 0.5), y, h);
}

void GridFunction::dump() const
{
	int idx = 0;
	
	for (int y = 0 ; y < m_numY ; y++)
	{
		for (int x = 0 ; x < m_numX ; x++, idx++)
		{
			std::cout << x << " " << y << " " << m_pValues[idx] << std::endl;
		}
		std::cout << std::endl;
	}
}

} // end namespace

