/*

  This file is a part of GRALE, a library to facilitate the simulation
  and inversion of gravitational lenses.

  Copyright (C) 2008-2012 Jori Liesenborgs

  Contact: jori.liesenborgs@gmail.com
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
*/

#include "graleconfig.h"
#include "sourceimage.h"
#include "circularsource.h"
#include "ellipticalsource.h"
#include "polygonsource.h"
#include "discretesource.h"
#include "pointsource.h"

#include "debugnew.h"

namespace grale
{

SourceImage::SourceImage(SourceType t,Vector2D<double> pos, double angle, double brightnessScale)
{
	SourceImage::parentplane = 0;
	SourceImage::pos = pos;
	SourceImage::angle = angle;
	m_brightnessScale = brightnessScale;
	stype = t;
	theta = (angle/180.0)*CONST_PI;
	id = 0;
}

SourceImage::~SourceImage()
{
	release();
}

bool SourceImage::attach(SourcePlane *p, uint16_t srcid)
{
	if (parentplane != 0)
	{
		setErrorString("Source is already attached to a source plane");
		return false;
	}
	if (p == 0)
	{
		setErrorString("Can't attach source to NULL");
		return false;
	}
	if (!p->registerSource(this,srcid))
	{
		setErrorString(std::string("Couldn't register source with source plane: ") + p->getErrorString());
		return false;
	}
	parentplane = p;
	id = srcid;
	return true;
}

bool SourceImage::release()
{
	if (parentplane == 0)
	{
		setErrorString("Source is not attached to a source plane");
		return false;
	}
	parentplane->unregisterSource(this);
	parentplane = 0;
	id = 0;
	return true;
}

#define SOURCEIMG_ID 			0x49435253
#define SOURCEIMG_TYPE_CIRCLE		1
#define SOURCEIMG_TYPE_ELLIPSE		2
#define SOURCEIMG_TYPE_DISCRETE		3
#define SOURCEIMG_TYPE_POLYGON		4
#define SOURCEIMG_TYPE_POINT		5

bool SourceImage::write(serut::SerializationInterface &si) const
{
	if (!isSerializable())
	{
		setErrorString("This source image type not serializable");
		return false;
	}

	int32_t typeID = 0;
	
	switch(stype)
	{
	case Circle:
		typeID = SOURCEIMG_TYPE_CIRCLE;
		break;
	case Ellipse:
		typeID = SOURCEIMG_TYPE_ELLIPSE;
		break;
	case Polygon:
		typeID = SOURCEIMG_TYPE_POLYGON;
		break;
	case Discrete:
		typeID = SOURCEIMG_TYPE_DISCRETE;
		break;
	case Point:
		typeID = SOURCEIMG_TYPE_POINT;
		break;
	default:
		setErrorString("Source image type not recognized");
		return false;
	}
	
	if (!si.writeInt32(SOURCEIMG_ID))
	{
		setErrorString(std::string("Couldn't write identifier: ") + si.getErrorString());
		return false;
	}

	if (!si.writeInt32(typeID))
	{
		setErrorString(std::string("Couldn't write source image type: ") + si.getErrorString());
		return false;
	}

	if (!si.writeDoubles(pos.getComponents(), 2))
	{
		setErrorString(std::string("Couldn't write source position: ") + si.getErrorString());
		return false;
	}

	if (!si.writeDouble(angle))
	{
		setErrorString(std::string("Couldn't write source angle: ") + si.getErrorString());
		return false;
	}

	if (!si.writeDouble(m_brightnessScale))
	{
		setErrorString(std::string("Couldn't write brightness scale factor: ") + si.getErrorString());
		return false;
	}
	
	if (!writeInternal(si))
		return false;
	
	return true;
}

bool SourceImage::read(serut::SerializationInterface &si,SourceImage **source,std::string &errstr)
{
	int32_t id,typeID;

	if (!si.readInt32(&id))
	{
		errstr = std::string("Couldn't read source image ID: ") + si.getErrorString();
		return false;
	}

	if (id != SOURCEIMG_ID)
	{
		errstr = std::string("Read invalid source image ID");
		return false;
	}

	if (!si.readInt32(&typeID))
	{
		errstr = std::string("Couldn't read source image type: ") + si.getErrorString();
		return false;
	}

	SourceImage *s = 0;
	
	switch(typeID)
	{
	case SOURCEIMG_TYPE_CIRCLE:
		s = new CircularSource();
		break;
	case SOURCEIMG_TYPE_ELLIPSE:
		s = new EllipticalSource();
		break;
	case SOURCEIMG_TYPE_POLYGON:
		s = new PolygonSource();
		break;
	case SOURCEIMG_TYPE_DISCRETE:
		s = new DiscreteSource();
		break;
	case SOURCEIMG_TYPE_POINT:
		s = new PointSource();
		break;
	default:
		errstr = std::string("Read invalid source image type");
		return false;
	}

	Vector2D<double> pos;
	double angle, theta, brightnessScale;
	
	if (!si.readDoubles(pos.getComponents(), 2))
	{
		delete s;
		errstr = std::string("Couldn't read source image position: ") + si.getErrorString();
		return false;
	}

	if (!si.readDouble(&angle))
	{
		delete s;
		errstr = std::string("Couldn't read source image angle: ") + si.getErrorString();
		return false;
	}

	if (!si.readDouble(&brightnessScale))
	{
		delete s;
		errstr = std::string("Couldn't read source brightness scale: ") + si.getErrorString();
		return false;
	}
	
	theta = (angle/180.0)*CONST_PI;
	
	s->pos = pos;
	s->angle = angle;
	s->theta = theta;
	s->m_brightnessScale = brightnessScale;
	
	if (!s->readInternal(si))
	{
		errstr = s->getErrorString();
		delete s;
		return false;
	}
	
	*source = s;
	
	return true;
}

bool SourceImage::writeInternal(serut::SerializationInterface &si) const
{
	setErrorString("Write not implemented");
	return false;
}

bool SourceImage::readInternal(serut::SerializationInterface &si)
{
	setErrorString("Read not implemented");
	return false;
}

} // end namespace

