/*

  This file is a part of GRALE, a library to facilitate the simulation
  and inversion of gravitational lenses.

  Copyright (C) 2008-2012 Jori Liesenborgs

  Contact: jori.liesenborgs@gmail.com
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
*/

#include "graleconfig.h"
#include "sourceplane.h"
#include "sourceimage.h"

#include "debugnew.h"

namespace grale
{

SourcePlane::SourcePlane(double D_s,double D_ds)
{
	SourcePlane::D_s = D_s;
	SourcePlane::D_ds = D_ds;
}

SourcePlane::~SourcePlane()
{
	std::map<uint16_t, SourceImage *>::const_iterator it;

	for (it = sources.begin() ; it != sources.end() ; it++)
	{
		// unregister the individual sources;
		(*it).second->parentplane = 0;
		(*it).second->id = 0;
	}
}

bool SourcePlane::registerSource(SourceImage *s, uint16_t sourceID)
{
	if (s == 0)
	{
		setErrorString("Source image can't be NULL");
		return false;
	}

	std::map<uint16_t, SourceImage *>::const_iterator it = sources.find(sourceID);

	if (it != sources.end())
	{
		setErrorString("Source identifier is already in use");
		return false;
	}
	sources[sourceID] = s;
	return true;
}

void SourcePlane::unregisterSource(SourceImage *s)
{
	std::map<uint16_t, SourceImage *>::iterator it;

	for (it = sources.begin() ; it != sources.end() ; it++)
	{
		if ((*it).second == s)
		{
			sources.erase(it);
			return;
		}
	}
}

double SourcePlane::getSurfaceBrightness(Vector2D<double> beta) const
{
	std::map<uint16_t, SourceImage *>::const_iterator it;
	double sum = 0;

	for (it = sources.begin() ; it != sources.end() ; it++)
	{
		const SourceImage *pSrc = (*it).second;
		
		if (pSrc->getSourceType() != SourceImage::Point) // Don't consider point sources here
			sum += (*it).second->getIntensity(beta);
	}
	return sum;
}

void SourcePlane::getPointSourceIntensities(Triangle2D<double> area, std::list<Vector2DPlus> &pointSourceInfo) const
{
	std::map<uint16_t, SourceImage *>::const_iterator it;

	pointSourceInfo.clear();
	
	for (it = sources.begin() ; it != sources.end() ; it++)
	{
		const SourceImage *pSrc = (*it).second;
		
		if (pSrc->getSourceType() == SourceImage::Point) // Only consider point sources here
		{
			Vector2D<double> position = pSrc->getAngularPosition();

			if (area.isInside(position))
				pointSourceInfo.push_back(Vector2DPlus(position.getX(), position.getY(), pSrc->getIntensity(position)));
		}
	}
}

bool SourcePlane::isSourceInRange(Vector2D<double> beta, double radius) const
{
	std::map<uint16_t, SourceImage *>::const_iterator it;

	for (it = sources.begin() ; it != sources.end() ; it++)
		if ((*it).second->isSourceInRange(beta, radius))
			return true;

	return false;
}

#define SOURCEPLANEID		0x414C5053

bool SourcePlane::write(serut::SerializationInterface &si,bool writesources /* = false */) const
{
	int32_t numsources = 0;
	double array[2];
	std::map<uint16_t, SourceImage *>::const_iterator it;
	
	if (!si.writeInt32(SOURCEPLANEID))
	{
		setErrorString(std::string("Error writing source plane ID: ")+si.getErrorString());
		return false;
	}
	
	array[0] = D_s;
	array[1] = D_ds;
	if (!si.writeDoubles(array,2))
	{
		setErrorString(std::string("Error writing D_s and D_ds: ")+si.getErrorString());
		return false;
	}
	
	if (writesources)
		numsources = sources.size();
	
	if (!si.writeInt32(numsources))
	{
		setErrorString(std::string("Error writing number of sources: ")+si.getErrorString());
		return false;
	}
	
	if (writesources)
	{
		for (it = sources.begin() ; it != sources.end() ; it++)
		{
			int32_t id = (int32_t)((*it).first);
			if (!si.writeInt32(id))
			{
				setErrorString(std::string("Error writing source ID: ")+si.getErrorString());
				return false;
			}
			if (!(*it).second->write(si))
			{
				setErrorString(std::string("Error writing source info: ")+(*it).second->getErrorString());
				return false;
			}
		}
	}

	return true;
}

bool SourcePlane::read(serut::SerializationInterface &si,SourcePlane **sp,std::string &errstr,bool loadsources)
{
	double array[2];
	int32_t numsources,i,id;

	if (!si.readInt32(&id))
	{
		errstr = std::string("Error reading source plane ID: ")+si.getErrorString();
		return false;
	}
	if (id != SOURCEPLANEID)
	{
		errstr = std::string("Read invalid source plane ID");
		return false;
	}
	if (!si.readDoubles(array,2))
	{
		errstr = std::string("Error reading D_s and D_ds: ")+si.getErrorString();
		return false;
	}
	if (!si.readInt32(&numsources))
	{
		errstr = std::string("Error reading number of sources: ")+si.getErrorString();
		return false;
	}

	SourcePlane *plane = new SourcePlane(array[0],array[1]);

	// Note: even if we do not wish to load the sources, we have to read them from the
	// stream anyway, to avoid running into problems when another object starts reading
	// from the stream.

	for (i = 0 ; i < numsources ; i++)
	{
		int32_t id;

		if (!si.readInt32(&id))
		{
			errstr = std::string("Error reading source ID: ")+si.getErrorString();
			return false;
		}
		
		SourceImage *src;
		std::string errstr2;

		if (!SourceImage::read(si,&src,errstr2))
		{
			std::map<uint16_t, SourceImage *>::const_iterator it;
			
			for (it = plane->sources.begin() ; it != plane->sources.end() ; it++)
			{
				(*it).second->parentplane = 0;
				delete (*it).second;
			}
			delete plane;
			errstr = std::string("Error reading source info: ") + errstr2;
			return false;
		}

		if (loadsources)
		{
			uint16_t srcID = (uint16_t)id;

			std::map<uint16_t, SourceImage *>::const_iterator it = plane->sources.find(srcID);

			if (it != plane->sources.end())
			{
				std::map<uint16_t, SourceImage *>::const_iterator it;
			
				for (it = plane->sources.begin() ; it != plane->sources.end() ; it++)
				{
					(*it).second->parentplane = 0;
					delete (*it).second;
				}
				delete plane;
				delete src;
				errstr = std::string("Detected duplicated source ID");
				return false;
			}
	
			plane->sources[srcID] = src;
			src->parentplane = plane;
			src->id = srcID;
		}
		else
			delete src;	
	}

	*sp = plane;
	
	return true;
}

} // end namespace
